#include "order_generator_random.h"

#include "cost_partitioning_heuristic.h"
#include "saturated_cost_partitioning_heuristic.h"
#include "utils.h"

#include "../option_parser.h"
#include "../plugin.h"

#include "../utils/countdown_timer.h"
#include "../utils/logging.h"
#include "../utils/rng.h"

using namespace std;

namespace cost_saturation {
OrderGeneratorRandom::OrderGeneratorRandom(const Options &opts) :
    OrderGenerator(opts),
    time_per_state(opts.get<double>("time_per_state")) {
}

void OrderGeneratorRandom::initialize(
    const Abstractions &abstractions,
    const vector<int> &) {
    utils::Log() << "Initialize random order generator" << endl;
    random_order = get_default_order(abstractions.size());
}

Order OrderGeneratorRandom::compute_order_for_state(
    const Abstractions &abstractions,
    const vector<int> &costs,
    const vector<int> &abstract_state_ids,
    bool) {
    if (time_per_state == 0.0) {
        rng->shuffle(random_order);
        return random_order;
    }
    utils::CountdownTimer timer(time_per_state);
    Order best_order;
    int max_h = -1;
    while (max_h == -1 || !timer.is_expired()) {
        rng->shuffle(random_order);
        CostPartitioningHeuristic cp = compute_saturated_cost_partitioning(
            abstractions, random_order, costs);
        int h = cp.compute_heuristic(abstract_state_ids);
        assert(h != INF);
        if (h > max_h) {
            best_order = random_order;
            max_h = h;
        }
    }
    assert(max_h >= 0);
    return best_order;
}


static shared_ptr<OrderGenerator> _parse_greedy(OptionParser &parser) {
    add_common_order_generator_options(parser);
    parser.add_option<double>(
        "time_per_state",
        "until hitting this time limit, compute multiple orders for a given "
        "state and return the best one. If the time limit is 0 (default), "
        "consider a single order.",
        "0.0",
        Bounds("0.0", "infinity"));
    Options opts = parser.parse();
    if (parser.dry_run())
        return nullptr;
    else
        return make_shared<OrderGeneratorRandom>(opts);
}

static Plugin<OrderGenerator> _plugin_greedy("random_orders", _parse_greedy);
}
